"""Module generator related to pooling operations.

- Author: Jongkuk Lim
- Contact: lim.jeikei@gmail.com
"""

from typing import Tuple, Union

import torch
from torch import nn

from kindle.modules.conv import Conv


class GlobalAvgPool(nn.AdaptiveAvgPool2d):
    """Global average pooling module.

    Arguments: []
    """

    def __init__(self):
        """Initialize."""
        super().__init__(output_size=1)


class SPP(nn.Module):
    """Spatial pyramid pooling layer used in YOLOv3-SPP.

    Arguments: [channel, [kernel_size1, kernel_size2, ...], activation]
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_sizes: Tuple[int, ...] = (5, 9, 13),
        activation: Union[str, None] = "ReLU",
    ) -> None:
        """Initialize Spatical Pyramid Pooling module.

        Args:
            in_channels: number of incoming channels
            out_channels: number of outgoing channels
            kernel_sizes: kernel sizes to use
            activation: Name of the activation to use in convolution.
        """
        super().__init__()
        hidden_channels = in_channels // 2
        self.conv1 = Conv(in_channels, hidden_channels, 1, 1, activation=activation)
        self.conv2 = Conv(
            hidden_channels * (len(kernel_sizes) + 1),
            out_channels,
            1,
            1,
            activation=activation,
        )
        self.pooling_modules = nn.ModuleList(
            [
                nn.MaxPool2d(kernel_size=k_size, stride=1, padding=k_size // 2)
                for k_size in kernel_sizes
            ]
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Feed forward."""
        x = self.conv1(x)
        return self.conv2(
            torch.cat([x] + [module(x) for module in self.pooling_modules], 1)
        )



"""Utilities for Kindle Model.

- Author: Jongkuk Lim
- Contact: lim.jeikei@gmail.com
"""
from __future__ import annotations

import os
import time
from typing import (TYPE_CHECKING, Any, Callable, Dict, List, Optional, Tuple,
                    Union)

import numpy as np
import torch
from ptflops import get_model_complexity_info
from tabulate import tabulate
from torch import nn
from tqdm import tqdm

if TYPE_CHECKING:
    from kindle.generator.base_generator import GeneratorAbstract


class ModelProfiler:
    """Model time consumption profiler."""

    def __init__(
        self,
        model: nn.Module,
        n_run: int = 100,
        input_size: Tuple[int, int] = (128, 128),
        batch_size: int = 1,
    ) -> None:
        """Initialize model profiler.

        Args:
            model: kindle.Model instance for profiling.
            n_run: number of inference to run.
            input_size: input size to test.
                    If model config contains "input_size", this will be ignored.
            batch_size: batch size to test.
        """
        self.model = model
        self.n_run = n_run
        self.input_size: Tuple[int, ...] = (
            batch_size,
            self.model.model_parser.cfg["input_channel"],  # type: ignore
        )

        if "input_size" in self.model.model_parser.cfg:  # type: ignore
            self.input_size += tuple(self.model.model_parser.cfg["input_size"])  # type: ignore
        else:
            self.input_size += input_size

        self.profile_result = [
            {"name": module.name, "time": np.zeros(self.n_run)}
            for module in self.model.model  # type: ignore
        ]
        self.n_running = 0

    @property
    def result_array(self) -> np.ndarray:
        """Profiling time result array."""
        return np.array([r["time"] for r in self.profile_result])

    @property
    def mean_run_time(self) -> float:
        """Mean time of the inference."""
        return float(self.result_array.sum(axis=0).mean())

    @property
    def std_run_time(self) -> float:
        """Standard deviation time of the inference."""
        return float(self.result_array.sum(axis=0).std())

    @property
    def total_run_time(self) -> float:
        """Total running time."""
        return float(self.result_array.sum())

    @property
    def sorted_index(self) -> np.ndarray:
        """Sorted indices by time consumption of the modules.

        Ex) First index element is the slowest module index.
        """
        return np.argsort(self.result_array.sum(axis=1))[::-1]

    @property
    def running_ratio(self) -> np.ndarray:
        """Running time consumption ratio of the modules."""
        result_array = self.result_array
        result = result_array.sum(axis=1) / result_array.sum()

        if isinstance(result, np.ndarray):
            result_out = result
        else:
            result_out = np.array(result)

        return result_out

    @torch.no_grad()
    def run(self, verbose: bool = True) -> None:
        """Run profiling.

        Args:
            verbose: print profiling result at the end.
        """
        self.profile_result = [
            {"name": module.name, "time": np.zeros(self.n_run)}
            for module in self.model.model  # type: ignore
        ]

        model_input = torch.zeros(self.input_size).to(
            list(self.model.parameters())[0].device
        )

        for run_idx in tqdm(range(self.n_run), desc="Profiling ..."):
            self.n_running = run_idx
            self.model.forward_once(model_input, profile_func=self._profile_func)  # type: ignore

        if verbose:
            self.print_result()

    def get_macs(self, verbose: bool = False) -> float:
        """Get MACs information using ptflops.

        This might not be correct as it is stated at
        (https://github.com/sovrasov/flops-counter.pytorch)

        Please use with caution.
        """
        macs, _ = get_model_complexity_info(
            model=self.model,
            input_res=self.input_size[1:],
            as_strings=False,
            print_per_layer_stat=verbose,
            verbose=verbose,
        )
        return macs

    def print_result(  # pylint: disable=too-many-locals
        self, sort_by_rank: bool = False
    ) -> None:
        """Print profiling result.

        Args:
            sort_by_rank: print sorted by time consumption rank.
        """

        print(f"Profiling result by {self.n_run:,} times running.", end="")
        if sort_by_rank:
            print(" Sorted by time consumption.")
        else:
            print(" Sorted by running order.")

        msg_title = (
            f"{'idx':>4} | {'Name':>20} | {'Time(Mean)':>10} | "
            f"{'Time(Std)':>10} | {'Time(Total)':>10} | "
            f"{'Rank'} | {'Ratio':>7} | {'Params':>13} |"
        )

        print("-" * len(msg_title))
        print(msg_title)
        print("-" * len(msg_title))

        slow_index = {idx: i for i, idx in enumerate(self.sorted_index)}
        running_ratio = self.running_ratio

        log_msgs = []

        for i, result in enumerate(self.profile_result):
            name = result["name"]
            time_mean = result["time"].mean()
            time_std = result["time"].std()
            time_sum = result["time"].sum()

            time_mean, t_unit = self._time_convert(time_mean)
            time_std, t_unit_std = self._time_convert(time_std)
            time_sum, t_unit_sum = self._time_convert(time_sum)
            log_msg = (
                f"{i:4d} | {name:>20} | {time_mean:7.2f} {t_unit:<2} | "  # type: ignore
                f"{time_std:7.2f} {t_unit_std:<2} | "
                f"{time_sum:8.2f} {t_unit_sum:<2} | "
                f"{slow_index[i]+1:4d} | {running_ratio[i]*100:6.2f}% | "
                f"{self.model.model[i].n_params:13,d} |"
            )
            log_msgs.append(log_msg)

        if sort_by_rank:
            loop: Union[np.ndarray, range] = self.sorted_index
        else:
            loop = range(len(log_msgs))

        for i in loop:
            print(log_msgs[i])

        total_time_mean, t_unit_mean = self._time_convert(self.mean_run_time)
        total_time_std, t_unit_std = self._time_convert(self.std_run_time)
        total_time_sum, t_unit_sum = self._time_convert(self.total_run_time)
        print("-" * len(msg_title))
        print(
            f"Running time\n"
            f" - Total : {total_time_sum:8.2f} {t_unit_sum:<2}\n"
            f" - {'Mean':>5} : {total_time_mean:8.2f} {t_unit_mean:<2}\n"
            f" - {'STD':>5} : {total_time_std:8.2f} {t_unit_std:<2}\n"
        )

    def get_parameter_numbers(self) -> int:
        """Get total number of parameters.

        Returns:
            Total number of parameters.
        """
        # TODO(jeikeilim): return the number of parameter list of each layers.
        n_param = sum([x.numel() for x in self.model.parameters()])
        return n_param

    @classmethod
    def _time_convert(cls, x: float) -> Tuple[float, str]:
        """Convert time units.

        Args:
            x: time to be converted (seconds).

        Returns:
            converted time value and its unit
            (seconds, milliseconds, microseconds, nanoseconds, picoseconds, femtoseconds)
        """
        time_units = ["s", "ms", "μs", "ns", "ps", "fs"]
        i = 0
        for i in range(len(time_units)):
            if x > 1.0:
                break
            x *= 1000

        return x, time_units[i]

    def _profile_func(self, module: Callable, x: torch.Tensor, i: int) -> torch.Tensor:
        """Profile callback function for kindle.Model.forward."""
        start_time = time.monotonic()
        y = module(x)
        time_took = time.monotonic() - start_time
        self.profile_result[i]["time"][self.n_running] = time_took

        return y


class ModelInfoLogger:
    """Kindle model information logger.

    This class is used once only on parsing the model.
    """

    def __init__(self, log_shapes: bool = False) -> None:
        """Initialize ModelInfoLogger instance."""
        self.log_shapes = log_shapes
        self.model_log_msg: List[Any] = []

    @property
    def head(self) -> List[str]:
        """Head message that contains column names."""
        log = [
            "idx",
            "from",
            "n",
            "params",
            "module",
            "arguments",
            "in_channel",
            "out_channel",
        ]
        if self.log_shapes:
            log.append("in_shape")
            log.append("out_shape")

        return log

    @property
    def info(self) -> str:
        """Model information containing head and messages."""
        return tabulate(self.model_log_msg, self.head, tablefmt="presto")

    def add(
        self,
        info: Tuple[int, int, int],
        module: nn.Module,
        module_generator: "GeneratorAbstract",
        args: List[Any],
        kwargs: Optional[Dict[str, Any]] = None,
        in_size: Optional[Union[np.ndarray, List]] = None,
        out_size: Optional[List[int]] = None,
    ) -> None:
        """Add module logging information.

        Args:
            info: (i, idx, repeat) Current parsing information.
            module: Parsed module.
            module_generator: Module generator used to generate module.
            args: Arguments of the module.
            kwargs: Keyword arguments of the module.
            in_size: Input size of the module.
                    Only required when {self.input_size} is not None.
            out_size: Output size of the module
                    Only required when {self.input_size} is not None.
        """
        i, idx, repeat = info

        args = args.copy()
        if module.name == "YamlModule":
            args[0] = args[0].split(os.sep)[-1].split(".")[0]

        args_str = str(args)
        if kwargs is not None:
            args_str += ", "
            for key, val in kwargs.items():
                args_str += f"{key}: {val}, "

            args_str = args_str[:-2]

        log = [
            f"{i:3d}",
            idx,
            repeat,
            f"{module.n_params:,d}",
            module.name,
            args_str,
            module_generator.in_channel,
            module_generator.out_channel,
        ]

        if self.log_shapes and in_size is not None and out_size is not None:
            in_size_str = str(in_size).replace("\n", ",")

            log.append(in_size_str)
            log.append(out_size)

        self.model_log_msg.append(log)



"""DWConv module.

- Author: Jongkuk Lim
- Contact: lim.jeikei@gmail.com
"""
import math
# pylint: disable=useless-super-delegation
from typing import Union

import torch
from torch import nn

from kindle.modules.activation import Activation
from kindle.utils.torch_utils import autopad


class DWConv(nn.Module):
    """Depthwise convolution with batch normalization and activation.

    Arguments: [channel, kernel_size, stride, padding, activation]
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        stride: int = 1,
        padding: Union[int, None] = None,
        activation: Union[str, None] = "ReLU",
    ) -> None:
        """Depthwise convolution with batch normalization and activation.

        Args:
            in_channels: input channels.
            out_channels: output channels.
            kernel_size: kernel size.
            stride: stride.
            padding: input padding. If None is given, autopad is applied
                which is identical to padding='SAME' in TensorFlow.
            activation: activation name. If None is given, nn.Identity is applied
                which is no activation.
        """
        super().__init__()
        # error: Argument "padding" to "Conv2d" has incompatible type "Union[int, List[int]]";
        # expected "Union[int, Tuple[int, int]]"
        self.conv = nn.Conv2d(
            in_channels,
            out_channels,
            kernel_size,
            stride,
            padding=autopad(kernel_size, padding),  # type: ignore
            groups=math.gcd(in_channels, out_channels),
            bias=False,
        )
        self.batch_norm = nn.BatchNorm2d(out_channels)
        self.activation = Activation(activation)()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward."""
        return self.activation(self.batch_norm(self.conv(x)))

    def fuseforward(self, x: torch.Tensor) -> torch.Tensor:
        """Fuse forward."""
        return self.activation(self.conv(x))



"""torch.nn module generator.

!!! Note
    Experimental feature.
    This might change in future release.
    Use with caution.

- Author: Jongkuk Lim
- Contact: lim.jeikei@gmail.com
"""
import inspect
from typing import Any, Dict, List

import numpy as np
import torch
from torch import nn

from kindle.generator.base_generator import GeneratorAbstract


class TorchNNModuleGenerator(GeneratorAbstract):
    """torch.nn module generator."""

    def __init__(self, *args, module_name: str, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.module_name = module_name
        self.base_module: nn.Module = getattr(
            __import__("torch.nn", fromlist=[""]), self.module_name[3:]
        )

    @property
    def out_channel(self) -> int:
        in_sizes = (self.in_channel, 128, 128)
        for i in range(len(in_sizes), 0, -1):
            try:
                out_shape = self.compute_out_shape(np.array(in_sizes[:i]))
                break
            except RuntimeError:
                continue
            except ValueError:
                continue

        return out_shape[0]

    @property
    def in_channel(self) -> int:
        return self.in_channels[self.from_idx]  # type: ignore

    @torch.no_grad()
    def compute_out_shape(self, size: np.ndarray, repeat: int = 1) -> List[int]:
        module = self(repeat=repeat)
        module.eval()
        module_out = module(torch.zeros([1, *list(size)]))

        return list(module_out.shape[1:])

    @property
    def kwargs(self) -> Dict[str, Any]:
        kwarg_name_0 = inspect.getfullargspec(self.base_module).args[1]
        if kwarg_name_0.startswith("in") or kwarg_name_0 in ["num_features"]:
            args = (self.in_channel, *self.args)
        else:
            args = self.args

        kwargs = self._get_kwargs(self.base_module, args)
        return kwargs

    def __call__(self, repeat: int = 1) -> nn.Module:
        if repeat > 1:
            module = [self.base_module(**self.kwargs) for _ in range(repeat)]
        else:
            module = self.base_module(**self.kwargs)

        out_module = self._get_module(module)
        out_module.name = self.module_name  # type: ignore
        return out_module



"""PyTorch Module Generator for parsing model yaml file.

- Author: Jongkuk Lim
- Contact: lim.jeikei@gmail.com
"""

from kindle.generator.add import AddGenerator
from kindle.generator.base_generator import GeneratorAbstract, ModuleGenerator
from kindle.generator.bottleneck import (BottleneckCSPGenerator,
                                         BottleneckGenerator)
from kindle.generator.concat import ConcatGenerator
from kindle.generator.conv import (ConvGenerator, DWConvGenerator,
                                   FocusGenerator)
from kindle.generator.custom_yaml_module import YamlModuleGenerator
from kindle.generator.flatten import FlattenGenerator
from kindle.generator.identity import IdentityGenerator
from kindle.generator.linear import LinearGenerator
from kindle.generator.nn import TorchNNModuleGenerator
from kindle.generator.poolings import (AvgPoolGenerator,
                                       GlobalAvgPoolGenerator,
                                       MaxPoolGenerator, SPPGenerator)
from kindle.generator.pretrained import (PreTrainedFeatureMapGenerator,
                                         PreTrainedGenerator)
from kindle.generator.upsample import UpSampleGenerator
from kindle.generator.yolo_head import YOLOHeadGenerator

__all__ = [
    "ModuleGenerator",
    "GeneratorAbstract",
    "BottleneckGenerator",
    "BottleneckCSPGenerator",
    "ConcatGenerator",
    "ConvGenerator",
    "DWConvGenerator",
    "FocusGenerator",
    "FlattenGenerator",
    "LinearGenerator",
    "AvgPoolGenerator",
    "GlobalAvgPoolGenerator",
    "MaxPoolGenerator",
    "SPPGenerator",
    "YamlModuleGenerator",
    "AddGenerator",
    "UpSampleGenerator",
    "IdentityGenerator",
    "TorchNNModuleGenerator",
    "PreTrainedGenerator",
    "PreTrainedFeatureMapGenerator",
    "YOLOHeadGenerator",
]



"""YOLOv5 head generator.

- Author: Jongkuk Lim
- Contact: lim.jeikei@gmail.com
"""

from typing import Any, Dict, List, Optional, Union

import numpy as np
from torch import nn

from kindle.generator.base_generator import GeneratorAbstract
from kindle.modules.yolo_head import YOLOHead


class YOLOHeadGenerator(GeneratorAbstract):
    """Generate YOLO Head."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.in_shape: Optional[Union[np.ndarray, List[List[int]]]] = None
        self.input_size: Optional[List[int]] = None

    @property
    def out_channel(self) -> List[int]:
        kwargs = self.kwargs
        n_layers = len(kwargs["anchors"])

        return [kwargs["n_classes"] + 5 for _ in range(n_layers)]

    @property
    def in_channel(self) -> List[int]:
        assert self.in_shape is not None

        return [s[0] for s in self.in_shape]

    def compute_out_shape(self, size: np.ndarray, repeat: int = 1) -> List[List[int]]:
        self.in_shape = size
        out_channels = self.out_channel
        return [[-1, o_channel] for o_channel in out_channels]

    @property
    def kwargs(self) -> Dict[str, Any]:
        kwargs = self._get_kwargs(YOLOHead, self.args)
        return kwargs

    def __call__(self, repeat: int = 1) -> nn.Module:
        assert self.in_shape is not None and self.input_size is not None
        in_shape = self.in_shape
        input_size = self.input_size
        kwargs = self.kwargs

        n_channels = [s[0] for s in in_shape]
        strides = [input_size[0] / s[1] for s in in_shape]

        kwargs["n_channels"] = n_channels
        kwargs["strides"] = strides

        module = YOLOHead(**kwargs)
        return self._get_module(module)



"""Module Description.

- Author: Jongkuk Lim
- Contact: lim.jeikei@gmail.com
"""

import setuptools

import kindle

with open("README.md", encoding="utf-8") as f:
    long_description = f.read()

with open("LICENSE", encoding="utf-8") as f:
    license_content = f.read().replace("\n", " ")

setuptools.setup(
    name="kindle",
    version=kindle.__version__,
    packages=setuptools.find_packages(),
    description="Kindle - Making a PyTorch model easier than ever!",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license=license_content,
    author="Jongkuk Lim",
    author_email="lim.jeikei@gmail.com",
    url="https://github.com/JeiKeiLim/kindle",
    download_url="",
    install_requires=[
        "tqdm>=4.56.0",
        "PyYAML>=5.3.1",
        "torch>=1.7.1",
        "ptflops>=0.6.4",
        "timm>=0.4.5",
        "tabulate>=0.8.9",
    ],
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    keyword="pytorch",
)
